;Lage der Positionen
;   0-----------1-----------2
;   |           |           |
;   |   8-------9------10   |
;   |   |       |       |   |
;   |   |  16--17--18   |   |
;   |   |   |       |   |   |
;   7--15--23      19--11---3
;   |   |   |       |   |   |
;   |   |  22--21--20   |   |
;   |   |       |       |   |
;   |  14------13------12   |
;   |           |           |
;   6-----------5-----------4

;Lage der Tripletts (3 in einer Reihe)
;   o----------------0---o-------------------o
;   |                    |                   |
;   |                    |                   |
;   |                    |                   |
;   |     o----------4---o-------------o     |
;   |     |              |             |     |
;   |     |             12             |     |
;   |     |              |             |     |
;   |     |     o----8---o-------o     |     |
;   |     |     |                |     |     |
;   |     |     |                9     5     1
;   |     |     |                |     |     |
;   o-----o--15-o                o-13--o-----o
;   |     |     |                |     |     |
;   3     7    11                |     |     |
;   |     |     |                |     |     |
;   |     |     o--------o--10---o     |     |
;   |     |              |             |     |
;   |     |             14             |     |
;   |     |              |             |     |
;   |     o--------------o---6---------o     |
;   |                    |                   |
;   |                    |                   |
;   |                    |                   |
;   o--------------------o---2---------------o

DeclareModule Brettstruktur
  UseModule Portierung

  ;"globale" Brettstrukturen und Verknpfungen
  ;die KI und deren Helferlein wird gegebenfalls in folgenden Arrays nachschauen und nach diesen Informationen suchen:
  ; -von welcher Position aus kann man sich wohin bewegen (also die direkten Nachbarpositionen)
  ; -eine bestimmte Position ist Bestandteil welcher Tripletts (dies sind immer zwei), oder ...
  ; -aus welchen (drei) Positionen setzt sich ein bestimmtes Triplett zusammen
  Structure BrettPosition
    BitMaske.q        ;enthlt die NachbarPositionen als BitMaske
    Array PIndex.i(4) ;enthlt die Indizes der NachbarPositionen in 4 Zugrichtungen: 0=hoch,1=rechts,2=unten,3=links
                      ;wenn eine Richtung den Wert -1 hat, ist ein Zug, vom Spielfeldlayout her, dorthin nicht mglich.
    Array TIndex.i(2) ;enthlt die Indizes der beiden Tripletts in denen sich die durch dieses structure definierte Position befindet
  EndStructure
  
  Structure Triplett
    BitMaske.q          ;enthlt die Positionen eines Tripletts als BitMaske
    Array PIndex.i(3)   ;enthlt die Indizes der Positionen eines Tripletts
    Orientierung.i      ;Orientierung des Triplett 0=waagerecht,1=senkrecht, derzeit nicht benutzt
  EndStructure

  Global Dim BrettPositionen.BrettPosition(23)
  Global Dim Tripletts.Triplett(15)
  Global EckPunkt.q,TPunkt.q,KreuzPunkt.q;Punkte mit strategischer Bedeutung
    
  ;mittlerweile maximale 86 Zugmglichkeiten gefunden*****************************
  #ZugSpeicherGroesse=100
EndDeclareModule


Module Brettstruktur
  UseModule Portierung

  Define.s csvzeile
  Define.i i
  EckPunkt  =%010101010101010101010101
  TPunkt    =%101010100000000010101010
  KreuzPunkt=%000000001010101000000000
  
  If ReadFile(0,"dat\BrettDaten.csv")
    ;Header weglesen
      csvzeile = ReadString(0)
    For i=0 To 23
      csvzeile = ReadString(0)
      BrettPositionen(i)\BitMaske  =Val( StringField(csvzeile, 1, ",") )
      BrettPositionen(i)\PIndex(0) =Val( StringField(csvzeile, 2, ",") )
      BrettPositionen(i)\PIndex(1) =Val( StringField(csvzeile, 3, ",") )
      BrettPositionen(i)\PIndex(2) =Val( StringField(csvzeile, 4, ",") )
      BrettPositionen(i)\PIndex(3) =Val( StringField(csvzeile, 5, ",") )
      BrettPositionen(i)\TIndex(0) =Val( StringField(csvzeile, 6, ",") )
      BrettPositionen(i)\TIndex(1) =Val( StringField(csvzeile, 7, ",") )
    Next i
    CloseFile(0)
  Else
    MessageRequester("Warn", "BrettDaten.csv", #PB_MessageRequester_Ok)
  EndIf
  
  If ReadFile(0,"dat\TriplettDaten.csv") 
    ;Header weglesen
      csvzeile = ReadString(0)
    For i=0 To 15
      csvzeile = ReadString(0)
      Tripletts(i)\BitMaske     =Val( StringField(csvzeile, 1, ",") )
      Tripletts(i)\PIndex(0)    =Val( StringField(csvzeile, 2, ",") )
      Tripletts(i)\PIndex(1)    =Val( StringField(csvzeile, 3, ",") )
      Tripletts(i)\PIndex(2)    =Val( StringField(csvzeile, 4, ",") )
      Tripletts(i)\Orientierung =Val( StringField(csvzeile, 5, ",") )
    Next i
    CloseFile(0)
  Else
    MessageRequester("Warn", "TriplettDaten.csv", #PB_MessageRequester_Ok)
  EndIf
EndModule


;****************************************************************************************
;****************************************************************************************
;****************************************************************************************


;Da bei Mhle ja eigentlich die Farben der Steine uninteressant sind (nur verschieden sollten sie sein),
;ein paar Erklrungen dazu:
;WEISS steht fr ANZIEHEND, also den ersten (Halb-)Zug machend
;SCHWARZ steht fr NACHZIEHEND, also den darauf folgenden (Halb-)Zug machend
;etliche Werte stehen in Arrays, in der Regel entspricht der Index 0 fr WEISS und der Index 1 fr SCHWARZ

;Lage der Positionen
;   0-----------1-----------2
;   |           |           |
;   |   8-------9------10   |
;   |   |       |       |   |
;   |   |  16--17--18   |   |
;   |   |   |       |   |   |
;   7--15--23      19--11---3
;   |   |   |       |   |   |
;   |   |  22--21--20   |   |
;   |   |       |       |   |
;   |  14------13------12   |
;   |           |           |
;   6-----------5-----------4
;KI-internes Speicherformat fr das Spielbrett:
;bergebene Brettdaten sind immer als "abgeschlossene Situation" oder Vergangenheit zu betrachten.

;soll heissen: der vollzogene Halbzug x hat jene Brettsituation hervorgebracht, die in BF_Brett() beschrieben ist!!!

;Vom Wert Halbzug kann nun folgendes abgeleitet werden:
;0          -> es wurde noch kein Zug gettigt
;1,3,5,7....-> ungerader Halbzug, WEISS war(!) am Halbzug
;2,4,6,8....-> gerader Halbzug, SCHWARZ war(!) am Halbzug
;der akuelle,noch nicht gettigte oder noch zu berechnende Halbzug ist um eins hher (oha, wie berraschend)
;die Farbe dafr wiederum ist nach den bekannten Regeln zu ermitteln
;Spielphasenbergang Setzphase->Schiebephase durch Halbzug exakt definiert:
;18->letzter Halbzug in Setzphase von SCHWARZ, 19->erster Halbzug in Schiebephase von WEISS

;BF_Brett() jedes gesetzte Bit entspricht eines gesetzten Steines auf dem Brett
;getrennt nach Farbe (BF_Brett(0) enthlt WEISS, BF_Brett(1) enthlt SCHWARZ)
;ein Beispiel:
;folgendes Brett W=weiss,S=schwarz
; W-----W-----W
; | +---+---+ |
; | | +-+-+ | |
; +-+-+   S-S-S
; | | +-+-+ | |
; | +---+---+ |
; +-----+-----+
;entspricht diesen Bitfeldern in BF_Brett():
;BF_Brett(0):000000000000000000000111
;BF_Brett(1):000010000000100000001000

DeclareModule Brett
  UseModule Portierung
  UseModule BrettStruktur

  Structure BrettDaten
    Halbzug.i
    Array BF_Brett.q(1)  
  EndStructure

  Declare.i ZaehleBits (Bitfeld.q)
  Declare.q PositionenFrei (*Quelle.BrettDaten)
  Declare.q PositionenBesetzt(*Quelle.BrettDaten)
  Declare.q NachbarFrei (*Quelle.BrettDaten, position.i)
  Declare.q SteineBeweglich(*Quelle.BrettDaten, SpielerIndex.i)
  Declare.q SteineGeschuetzt(*Quelle.BrettDaten, SpielerIndex.i)
  Declare.q SteineUngeschuetzt(*Quelle.BrettDaten, SpielerIndex.i)
  Declare.i PositionInMuehle (*Quelle.BrettDaten, position.i)
  Declare.i AktiverSpieler (*Quelle.BrettDaten)
  Declare.i PassiverSpieler (*Quelle.BrettDaten)
  Declare.i AktuelleSpielPhase(*Quelle.BrettDaten)
  Declare.i SteineInTriplett (*Quelle.BrettDaten,TriplettIndex.i,SpielerIndex.i)
  Declare.i BrettBewertung (*Quelle.BrettDaten, Tiefe.i)
EndDeclareModule


Module Brett
  UseModule Portierung
  UseModule Brettstruktur
    
  ;zhlt die Bits eines Bitfeldes(max 24 Bits)
  Procedure.i ZaehleBits (Bitfeld.q)
    Define.i zaehler,i
    zaehler=0
    For i=0 To 23
      If bit(Bitfeld,i) 
        zaehler=zaehler+1
      EndIf
    Next i
    ProcedureReturn zaehler
  EndProcedure

  
  ;liefert alle freien Positionen des Spielfeldes als Bitfeld
  Procedure.q PositionenFrei (*Quelle.BrettDaten)
    ProcedureReturn ~(*Quelle\BF_Brett(0) | *Quelle\BF_Brett(1)) & BitLimitter
  EndProcedure
 
  
  ;liefert alle besetzten Positionen des Spielfeldes
  Procedure.q PositionenBesetzt(*Quelle.BrettDaten)
     ProcedureReturn (*Quelle\BF_Brett(0) | *Quelle\BF_Brett(1))
  EndProcedure
  
  
  ;liefert die freien Nachbarpositionen relativ zu einer bestimmten Brettposition
  Procedure.q NachbarFrei (*Quelle.BrettDaten, position.i)
     ProcedureReturn (PositionenFrei(*Quelle) & BrettPositionen(position)\BitMaske)
  EndProcedure
  
  
  ;liefert alle Steine einer Farbe, die sich bewegen knnen 
  Procedure.q SteineBeweglich(*Quelle.BrettDaten, SpielerIndex.i)
    Define.i frei,auflaufend
    Define.i position
    frei=PositionenFrei(*Quelle)
    auflaufend=0
    For position=0 To 23
      If Bit(*Quelle\BF_Brett(SpielerIndex),position)
        ;sollten im Ergebnis der AND Verkpfung "freie Brettpositionen" mit "Nachbarpositionen" Bits gesetzt sein,
        ;so kann mann sagen, ein auf dieser Position befindlicher Stein lt sich bewegen... wohin ist hier erstmal egal
        If (frei & BrettPositionen(position)\BitMaske)
          auflaufend=BitSet(auflaufend,position) 
        EndIf
      EndIf
    Next position
    ProcedureReturn auflaufend
  EndProcedure
  
  
  ;liefert alle geschtzten Steine einer Farbe
  Procedure.q SteineGeschuetzt(*Quelle.BrettDaten, SpielerIndex.i)
    Define.q tmp,auflaufend
    Define.i i
    auflaufend=0
    For i.i=0 To 15
      ;mit folgendem enthlt tmp alle Steine von SpielerIndex die innerhalb des Tripletts(i) liegen. Das knnen drei sein, mssen aber nicht
      tmp= *Quelle\BF_Brett(SpielerIndex) & Tripletts(i)\BitMaske
      If tmp = Tripletts(i)\BitMaske;wenn die gleichen Bits wie im Triplett selbst gesetzt sind, also in beiden jeweils drei dann...
        auflaufend = auflaufend | tmp;in ;auflaufend; alle Bits von tmp sammeln
      EndIf
    Next i
    ProcedureReturn auflaufend
  EndProcedure
  
  
  ;Funktion liefert alle ungeschtzten Steine eines Spielers
  Procedure.q SteineUngeschuetzt(*Quelle.BrettDaten, SpielerIndex.i)
    Define.q geschuetzt
    geschuetzt=SteineGeschuetzt(*Quelle, SpielerIndex)
    ProcedureReturn *Quelle\BF_Brett(SpielerIndex) &  (~geschuetzt)
  EndProcedure
  
  ;Primr dient diese Funktion dazu festzustellen, ob eine gerade besetzte Position zum Schliessen keiner/einer/zweier Mhle(n) gefhrt hat.
  Procedure.i PositionInMuehle (*Quelle.BrettDaten, position.i)
    Dim TriplettInhalt.q(1)
    Define.i SpielerIndex,gefunden
    gefunden=0
    ;mit welcher Farbe ist die zu untersuchende Position besetzt, denn nur diese hat ja berhaupt die Mglichkeit eine Mhle zu bilden
    ;es kann nur eine der beiden sein, aber selbst das ist kein muss, dann ist das Ergebnis auf jeden Fall 0
    ;wenn es WEISS nicht ist, wird einfach angenommen dass es dann SCHWARZ ist, fertig:)
    If Bit(*Quelle\BF_Brett(0),position)
      SpielerIndex=0
    Else                                    
      SpielerIndex=1
    EndIf
    
    
    TriplettInhalt(0)= *Quelle\BF_Brett(SpielerIndex) & Tripletts( Brettpositionen(position)\TIndex(0) )\BitMaske
    TriplettInhalt(1)= *Quelle\BF_Brett(SpielerIndex) & Tripletts( Brettpositionen(position)\TIndex(1) )\BitMaske
    
    If ( TriplettInhalt(0) = Tripletts( Brettpositionen(position)\TIndex(0) )\BitMaske )
      gefunden=gefunden+1
    EndIf
    If ( TriplettInhalt(1) = Tripletts( Brettpositionen(position)\TIndex(1) )\BitMaske )
      gefunden=gefunden+1
    EndIf
  
     ProcedureReturn gefunden
  EndProcedure
  
  
  ;ermittelt die aktuelle Spielfarbe anhand bergebener Brettdaten
  Procedure.i AktiverSpieler (*Quelle.BrettDaten)
    Define.i result
    If ((*Quelle\Halbzug + 1) % 2)=0
      result=1
    Else
      result=0    
    EndIf
    ProcedureReturn result;bei geradem Halbzug, also x mod 2=0, Schwarz sonst Weiss
  EndProcedure
  
  
  Procedure.i PassiverSpieler (*Quelle.BrettDaten)
    Define.i result
    If ((*Quelle\Halbzug + 2) % 2)=0
      result=1
    Else
      result=0    
    EndIf
    ProcedureReturn result;bei geradem Halbzug, also x mod 2=0, Schwarz sonst Weiss
  EndProcedure
  
  
  ;ermittelt die aktuelle SpielPhase anhand bergebener Brettdaten
  Procedure.i AktuelleSpielPhase(*Quelle.BrettDaten)
    Define.i Phase, steine,Halbzug,SpielerIndex
    Halbzug =*Quelle\Halbzug + 1;der noch auszufhrende Halbzug ist um eins hher als der des bergebenen Spielbretts
    SpielerIndex=AktiverSpieler(*Quelle)
    If Halbzug<19
      Phase=1           ;Halbzug 1 bis 18 definitiv SetzPhase
    Else                ;danach entscheidet die Anzahl der Steine des Spielers ber Phase
      steine=ZaehleBits(*Quelle\BF_Brett(SpielerIndex));Anzahl der Steine der betreffenden Farbe zhlen
      If steine>3
        Phase=2  ;sind mehr als drei Steine da, dann SchiebePhase
      EndIf
      If steine=3
        Phase=3  ;bei drei Steinen SpringPhase
      EndIf
      If steine<3
        Phase=0  ;bei uneindeutiger Brettsituation (KI Test)
      EndIf
    EndIf
    ProcedureReturn Phase
  EndProcedure
  
  
  Procedure.i SteineInTriplett (*Quelle.BrettDaten,TriplettIndex.i,SpielerIndex.i)
    Define.i i,zaehler=0
    For i=0 To 2
      If bit(*Quelle\BF_Brett(SpielerIndex) , Tripletts(TriplettIndex)\PIndex(i) )
        zaehler=zaehler+1
      EndIf
    Next i
     ProcedureReturn zaehler
  EndProcedure
  
  
  ;Bewertet werden bestimmte (genauer: acht) Mengenanteile der Steine eines Tripletts, die Lage der Steine innerhalb ist hier vllig uninteressant
  ;Brettbewertung kann sowohl positiv als auch negativ sein.
  ;Je HHER (+) der Wert ist, desto mehr ist WEISS im Vorteil,
  ;je KLEINER (-) der Wert ist, desto mehr ist SCHWARZ im Vorteil.
  ;ab +1024 und grer ist das Brett fr WEISS gewonnen
  ;ab -1024 und kleiner ist das Brett fr SCHWARZ gewonnen  
  Procedure.i BrettBewertung (*Quelle.BrettDaten, Tiefe.i)
    
    Define.i auflaufend, bewertung,TriplettIndex, i
    Define.q TriplettInhalt  
    Dim EckPunktZaehler.i(1)
    Dim TPunktZaehler.i(1)
    Dim KreuzPunktZaehler.i(1)
    Dim Steinzaehler.i(1)
    Dim TriplettSteinZaehler.i(1)
    Dim EckPunkte.q(1)
    Dim TPunkte.q(1)
    Dim KreuzPunkte.q(1)
    Dim beweglich.q(1)
  
    Steinzaehler(0)=0
    Steinzaehler(1)=0
    beweglich(0)=0
    beweglich(1)=0
  
    ;Bitfelder erzeugen
    EckPunkte(0)=*Quelle\BF_Brett(0) & EckPunkt
    EckPunkte(1)=*Quelle\BF_Brett(1) & EckPunkt
    TPunkte(0)=*Quelle\BF_Brett(0) & TPunkt
    TPunkte(1)=*Quelle\BF_Brett(1) & TPunkt
    KreuzPunkte(0)=*Quelle\BF_Brett(0) & KreuzPunkt
    KreuzPunkte(1)=*Quelle\BF_Brett(1) & KreuzPunkt
    ;Beschleuniger, nur eine FOR...NEXT zum gleichzeitigen Durchzhlen einfacher Bitmasken
    For i=0 To 23
      If bit(EckPunkte(0),i)
        EckPunktZaehler(0) =EckPunktZaehler(0)+1;Eckpunkte weiss zhlen
      EndIf
      If bit(EckPunkte(1),i)
        EckPunktZaehler(1) =EckPunktZaehler(1)+1;Eckpunkte schwarz zhlen
      EndIf
      If bit(TPunkte(0),i)
        TPunktZaehler(0) =TPunktZaehler(0)+1;Tpunkte weiss zhlen
      EndIf    
      If bit(TPunkte(1),i)
        TPunktZaehler(1) =TPunktZaehler(1)+1;Tpunkte schwarz zhlen
      EndIf
      If bit(KreuzPunkte(0),i)
        KreuzPunktZaehler(0) =KreuzPunktZaehler(0)+1;Kreuzpunkte weiss zhlen
      EndIf
      If bit(KreuzPunkte(1),i)
        KreuzPunktZaehler(1) =KreuzPunktZaehler(1)+1;Kreuzpunkte schwarz zhlen
      EndIf
      If bit(*Quelle\BF_Brett(0),i)
        Steinzaehler(0) =Steinzaehler(0)+1;Steine weiss zhlen
      EndIf
      If bit(*Quelle\BF_Brett(1),i)
        Steinzaehler(1) =Steinzaehler(1)+1;Steine schwarz zhlen    
      EndIf
    Next i
  
    If *Quelle\Halbzug>18;Abbruchbedingung fr Schiebe- u. Springphase, in Setzphase kann das Spiel niemals verloren werden
      beweglich(0)=SteineBeweglich(*Quelle, 0)
      beweglich(1)=SteineBeweglich(*Quelle, 1)
      
      ;WEISS hat gewonnen, weil ...
      If Steinzaehler(1)<3;Schwarz weniger als 3 Steine
        bewertung= 1024
      EndIf
      If (beweglich(1)=0) And (Steinzaehler(1)>3);Schwarz unbeweglich
        bewertung= 1024
      EndIf  
      
      ;SCHWARZ hat gewonnen, weil ...
      If Steinzaehler(0)<3 ;Weiss weniger als 3 Steine
        bewertung= -1024
      EndIf  
      If (beweglich(0)=0) And (Steinzaehler(0)>3);Weiss unbeweglich
        bewertung= -1024 
      EndIf
      If bewertung<>0;ermglicht dem KI-Algo zu unterscheiden, ob ein Gewinnzug direkt im nchsten Zug mglich oder erst tiefer im Suchbaum zu finden ist
        bewertung =bewertung*(Tiefe+1);"pimp the Abbruchbedingung with the Tiefe" :)
      EndIf
    
    EndIf
  
    ;!!!!!!!!!!!!!!!!!!!!!!!
    ;wenn bis hierher keine Bedingung fr ein Spielende vorliegt >>>>(bewertung=0)<<<<, dann erfolgt eine "normale" Bewertung des Brettes:
    ;!!!!!!!!!!!!!!!!!!!!!!!
  
    If (bewertung=0)
      For TriplettIndex=0 To 15
  
        TriplettSteinZaehler(0)=SteineInTriplett (*Quelle ,TriplettIndex,0)
        TriplettSteinZaehler(1)=SteineInTriplett (*Quelle ,TriplettIndex,1)
  
        If (TriplettSteinZaehler(0)=2) And (TriplettSteinZaehler(1)=0);ww-
          bewertung =bewertung +8
        EndIf
        
        If (TriplettSteinZaehler(0)=3) And (TriplettSteinZaehler(1)=0);www
          bewertung =bewertung +6
        EndIf
        
        If (TriplettSteinZaehler(0)=1) And (TriplettSteinZaehler(1)=2);wss !! Mehrheit liegt ja eigentlich beim Gegner aber Mhlengefahr gebannt, daher "Belohnung" fr weiss!!
          bewertung =bewertung +2
        EndIf
        
        If (TriplettSteinZaehler(0)=1) And (TriplettSteinZaehler(1)=0)
          bewertung =bewertung +1;w--
        EndIf
        
        ; ---  sw-  ws-  werden nicht oder als neutral gewertet
              
        If (TriplettSteinZaehler(0)=0) And (TriplettSteinZaehler(1)=1);s--
          bewertung =bewertung -1
        EndIf
        
        If (TriplettSteinZaehler(0)=2) And (TriplettSteinZaehler(1)=1);sww !! Mehrheit liegt ja eigentlich beim Gegner aber Mhlengefahr gebannt, daher "Belohnung" fr schwarz!!
          bewertung =bewertung -2
        EndIf
        
        If (TriplettSteinZaehler(0)=0) And (TriplettSteinZaehler(1)=3);sss
          bewertung =bewertung -6
        EndIf
        
        If (TriplettSteinZaehler(0)=0) And (TriplettSteinZaehler(1)=2);ss-
          bewertung =bewertung -4
        EndIf
        
      Next TriplettIndex
  
      ;Eckpunkte ungnstig(!?), keine Ahnung, hab ich irgendwo im WWW gelesen -> negative Bewertung
      bewertung =bewertung -EckPunktZaehler(0)
      bewertung =bewertung +EckPunktZaehler(1)
      bewertung =bewertung +TPunktZaehler(0)*2
      bewertung =bewertung -TPunktZaehler(1)*2
      bewertung =bewertung +KreuzPunktZaehler(0)*3
      bewertung =bewertung -KreuzPunktZaehler(1)*3
      ;Anzahl Steine einfliessen lassen
      bewertung =bewertung +SteinZaehler(0)*16
      bewertung =bewertung -SteinZaehler(1)*16
    EndIf
    ProcedureReturn bewertung
  EndProcedure
  
EndModule

; IDE Options = PureBasic 5.60 (Windows - x64)
; CursorPosition = 270
; FirstLine = 189
; Folding = ---
; EnableXP
; DisableDebugger